<?php

class gnupgReplacement {
  const GNUPG_ERROR_WARNING = 1;
  const GNUPG_ERROR_EXCEPTION = 2;
  const GNUPG_ERROR_SILENT = 3;
  const GNUPG_PROTOCOL_OpenPGP = 0;
  const GNUPG_PROTOCOL_CMS = 1;
  const GNUPG_SIG_MODE_NORMAL = 0;
  const GNUPG_SIG_MODE_DETACH = 1;
  const GNUPG_SIG_MODE_CLEAR = 2;

  private $gnupgDir;
  private $decryptKeys = array();
  private $encryptKeys = array();
  private $signKeys = array();
  private $signMode = self::GNUPG_SIG_MODE_NORMAL;
  private $errorMode = self::GNUPG_ERROR_SILENT;
  private $lastErr = '';

  public function __construct () {
    $this->gnupgDir = getenv('GNUPGHOME').'openpgp/';
    require_once dirname(__FILE__).'/openpgp_crypt_rsa.php';
    require_once dirname(__FILE__).'/openpgp_crypt_symmetric.php';
    if (!is_dir($this->gnupgDir))
      mkdir($this->gnupgDir);
  }

  private function err ($msg) {
    switch ($this->errorMode) {
      case self::GNUPG_ERROR_WARNING:
        trigger_error($msg, E_USER_WARNING);
        break;
      case self::GNUPG_ERROR_EXCEPTION:
        throw new Exception($msg);
        break;
      case self::GNUPG_ERROR_SILENT:
        $this->lastErr = $msg;
        break;
    }
  }

  private function addkey (&$arr, string $fingerprint, string $passphrase='') {
    if ($passphrase != '') {
      $this->err('Keys with passphrase are currently not supported!'); // TODO: Fix this.
      return false;
    }
    if (!is_file($this->gnupgDir.$fingerprint.'.gpg')) {
      $this->err('Key not found!');
      return false;
    }
    $keys = OpenPGP_Message::parse(file_get_contents($this->gnupgDir.$fingerprint.'.gpg'));
    $oldCount = count($arr);
    foreach ($keys as $keyPacket)
      $arr[] = $keyPacket;
    return count($arr) > $oldCount;
  }

  public function adddecryptkey (string $fingerprint, string $passphrase='') { // Add a key for decryption
    return $this->addkey($this->decryptKeys, $fingerprint, $passphrase);
  }

  public function addencryptkey (string $fingerprint) { // Add a key for encryption
    return $this->addkey($this->encryptKeys, $fingerprint);
  }

  public function addsignkey (string $fingerprint, string $passphrase = '') { // Add a key for signing
    return $this->addkey($this->signKeys, $fingerprint, $passphrase);
  }

  public function cleardecryptkeys () { // Removes all keys which were set for decryption before
    $this->decryptKeys = array();
    return true;
  }

  public function clearencryptkeys () { // Removes all keys which were set for encryption before
    $this->encryptKeys = array();
    return true;
  }

  public function clearsignkeys () { // Removes all keys which were set for signing before
    $this->signKeys = array();
    return true;
  }

  public function decrypt (string $text) { // Decrypts a given text
    return false; // TODO: implement!
  }

  public function decryptverify (string $text, string &$plaintext) { // Decrypts and verifies a given text
    return false; // TODO: implement!
  }

  public function encrypt (string $plaintext) { // Encrypts a given text
    if (count($this->encryptKeys) == 0) {
      $this->err('No encryption keys given!');
      return false;
    }
    $keys = new OpenPGP_Message();
    for ($i = 0; $i < count($this->encryptKeys); $i++)
      $keys[] = $this->encryptKeys[$i];
    try {
      $data = new OpenPGP_LiteralDataPacket($plaintext, array('format' => 'u', 'filename' => 'secret.eml'));
      $encrypted = OpenPGP_Crypt_Symmetric::encrypt($keys, new OpenPGP_Message(array($data)));
      $enc = OpenPGP::enarmor($encrypted->to_bytes(), 'PGP MESSAGE');
    } catch (Exception $e) {
      $this->err('Encryption failed');
      return false;
    }
    return $enc;
  }

  public function encryptsign (string $plaintext) { // Encrypts and signs a given text
    if (count($this->signKeys) == 0) { // TODO: Test also for multiple sign keys!
      $this->err('No sign keys given!');
      return false;
    }
    if (count($this->encryptKeys) == 0) {
      $this->err('No encryption keys given!');
      return false;
    }
    $ekeys = new OpenPGP_Message();
    for ($i = 0; $i < count($this->encryptKeys); $i++)
      $ekeys[] = $this->encryptKeys[$i];
    $skeys = new OpenPGP_Message();
    for ($i = 0; $i < count($this->signKeys); $i++)
      $skeys[] = $this->signKeys[$i];
    $enc = false;
    try {
      $data = new OpenPGP_LiteralDataPacket($plaintext, array('format' => 'u', 'filename' => 'secret.eml'));
      $signer = new OpenPGP_Crypt_RSA($skeys);
      $signed = $signer->sign($data);
      if (!$signed)
        $this->err('Signing failed');
      else {
        $encrypted = OpenPGP_Crypt_Symmetric::encrypt($ekeys, new OpenPGP_Message(array($signed)));
        $enc = OpenPGP::enarmor($encrypted->to_bytes(), 'PGP MESSAGE');
      }
    } catch (Exception $e) {
      $this->err('Encryption failed');
      return false;
    }
    return $enc;
  }

  public function export (string $fingerprint) { // Exports a key
    if (!is_file($this->gnupgDir.$fingerprint.'.gpg')) {
      $this->err('Key not found!');
      return false;
    }
    $keys = OpenPGP_Message::parse(file_get_contents($this->gnupgDir.$fingerprint.'.gpg'));
    $publicKeys = new OpenPGP_Message();
    foreach ($keys as $keyPacket) {
      if ($keyPacket instanceof OpenPGP_SecretSubkeyPacket)
        $publicKeys[] = new OpenPGP_PublicSubkeyPacket($keyPacket);
      else if ($keyPacket instanceof OpenPGP_SecretKeyPacket)
        $publicKeys[] = new OpenPGP_PublicKeyPacket($keyPacket);
      else
        $publicKeys[] = clone($keyPacket);
    }
    return OpenPGP::enarmor($publicKeys->to_bytes(), 'PGP PUBLIC KEY BLOCK');
  }

  public function geterror () { // Returns the errortext, if a function fails
    if ($this->lastErr == '' || $this->errorMode != self::GNUPG_ERROR_SILENT) return false;
    else return $this->lastErr;
  }

  public function getprotocol () { // Returns the currently active protocol for all operations
    return self::GNUPG_PROTOCOL_OpenPGP; // TODO: How can this ever switch to CMS?
  }

  public function import (string $keydata) { // Imports a key
    if (strpos($keydata, '-----BEGIN PGP PUBLIC KEY BLOCK-----') !== false)
      $gpgdata = OpenPGP::unarmor($keydata, 'PGP PUBLIC KEY BLOCK');
    else if (strpos($keydata, '-----BEGIN PGP PRIVATE KEY BLOCK-----') !== false)
      $gpgdata = OpenPGP::unarmor($keydata, 'PGP PRIVATE KEY BLOCK');
    else {
      $this->err('Unknown key type!');
      return false;
    }
    try {
      $keys = OpenPGP_Message::parse($gpgdata);
    } catch (Exception $e) {
      $this->err('Error parsing key data!');
      return false;
    }
    if (!$keys) {
      $this->err('Error parsing key data!');
      return false;
    }
    $info = array(
      'imported' => 0,
      'unchanged' => 0,
      'newuserids' => 0,
      'newsubkeys' => 0,
      'secretimported' => 0,
      'secretunchanged' => 0,
      'newsignatures' => 0,
      'skippedkeys' => 0,
      'fingerprint' => '',
    );
    foreach ($keys as $keyPacket)
      if ($keyPacket instanceof OpenPGP_PublicKeyPacket) {
        $fp = $keyPacket->fingerprint;
        if (!($keyPacket instanceof OpenPGP_PublicSubkeyPacket) && !($keyPacket instanceof OpenPGP_SecretSubkeyPacket)) {
          if ($info['fingerprint'] != '') { // TODO support files with multiple keys! - Currently the complete key file including all sub-keys is stored under the main key's fingerprint.
            $this->err('Only one key per file supported!');
            return false;
          }
          $info['fingerprint'] = $fp;
          if (is_file($this->gnupgDir.$fp.'.gpg')) $info['unchanged']++;
          else $info['imported']++;
          if ($keyPacket instanceof OpenPGP_SecretKeyPacket) {
            if (is_file($this->gnupgDir.$fp.'.gpg')) $info['secretunchanged']++;
            else $info['secretimported']++;
          }
          if (file_put_contents($this->gnupgDir.$fp.'.gpg', $gpgdata) === false) {
            $this->err('Error writing to key file!');
            return false;
          }
        }
      }
    return $info;
  }

  public function keyinfo (string $pattern) { // Returns an array with information about all keys that matches the given pattern
    return false; // TODO: implement!
  }

  public function setarmor (int $armor) { // Toggle armored output
    return false; // TODO: implement!
  }

  public function seterrormode (int $errormode) { // Sets the mode for error_reporting
    if ($errormode == self::GNUPG_ERROR_WARNING || $errormode == self::GNUPG_ERROR_EXCEPTION || $errormode == self::GNUPG_ERROR_SILENT)
      $this->errorMode = $errormode;
  }

  public function setsignmode (int $signmode) { // Sets the mode for signing
    if ($signmode == self::GNUPG_SIG_MODE_NORMAL || $signmode == self::GNUPG_SIG_MODE_DETACH || $signmode == self::GNUPG_SIG_MODE_CLEAR)
      $this->signMode = $signmode;
  }

  public function sign (string $plaintext) { // Signs a given text
    return false; // TODO: implement!
  }

  public function verify (string $signed_text, string $signature, string &$plaintext='') { // Verifies a signed text
    return false; // TODO: implement!
  }

}
